@@ -3,7 +3,11 @@ package PLS;
 use strict;
 use warnings;
 
-our $VERSION = '0.8';
+our $VERSION = '0.1';
+
+1;
+
+__END__
 
 =head1 NAME
 
@@ -93,7 +97,3 @@ Copyright 2021 Marc Reisner
 
 This library is free software; you may redistribute it and/or
 modify it under the same terms as Perl itself.
-
-=cut
-
-1;
